module type MAP_TYPE =
  sig
    type key
(*    type (+'a) t *)
    type (+ 'a) t
    val empty: 'a t
    val is_empty: 'a t -> bool
    val add: key -> 'a -> 'a t -> 'a t
    val find: key -> 'a t -> 'a
    val remove: key -> 'a t -> 'a t
    val mem:  key -> 'a t -> bool
    val iter: (key -> 'a -> unit) -> 'a t -> unit
    val iter_rev: (key -> 'a -> unit) -> 'a t -> unit
    val map: ('a -> 'b) -> 'a t -> 'b t
    val mapi: (key -> 'a -> 'b) -> 'a t -> 'b t
    val mapiz : (key -> 'a -> 'a) -> 'a t -> 'a t

    val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val map2iz: (key -> 'a -> 'a -> 'a) -> 'a t -> 'a t -> 'a t
    val map2i:  (key -> 'a -> 'a) -> (key -> 'a -> 'a) ->  
                (key -> 'a -> 'a -> 'a) -> 'a t -> 'a t -> 'a t
    val map2io: (key -> 'a -> 'a option) -> (key -> 'a -> 'a option) ->
                (key -> 'a -> 'a -> 'a option) -> 'a t -> 'a t -> 'a t
    val map2:   (key -> 'a -> 'c) -> (key -> 'b -> 'c) ->  
                (key -> 'a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
    val iter2iz: (key -> 'a -> 'a -> unit) -> ('a t -> 'a t -> unit)
    val iter2i: (key -> 'a -> unit) -> (key -> 'a -> unit) ->
	        (key -> 'a -> 'a -> unit) -> ('a t -> 'a t -> unit)
    val iter2: (key -> 'a -> unit) -> (key -> 'b -> unit) ->
	        (key -> 'a -> 'b -> unit) -> ('a t -> 'b t -> unit)
    val forall2iz: (key -> 'a -> 'a -> bool) -> 'a t -> 'a t -> bool
    val fold2: (key -> 'a -> 'c -> 'c) ->
       (key -> 'b -> 'c -> 'c) ->
       (key -> 'a -> 'b -> 'c -> 'c) -> 'a t -> 'b t -> 'c -> 'c
    val fold2i: (key -> 'a -> 'b -> 'c -> 'c) -> 'a t -> 'b t -> 'c -> 'c
    val equal: ('a -> 'a -> bool) -> 'a t -> 'a t -> bool

    type 'a delta_record
    val compute_delta: ('a->'a->bool) -> 'a t -> 'a t -> 'a delta_record
    val apply_delta: 'a t -> 'a delta_record -> 'a t

(*
    (* applies the function to all the tuples *)
    val map_two: (key  -> 'a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
    val fold_two: (key -> 'a -> 'b -> 'c -> 'c) -> 'a t -> 'b t -> 'c -> 'c
    val iter_two: (key -> 'a -> 'b -> unit) -> 'a t -> 'b t -> unit
*)

    val size: 'a t -> int

    val iter_slice: (key -> 'a -> unit) -> ('a t) -> key -> key -> unit
    val fold_slice: (key -> 'a -> 'b -> 'b) -> ('a t) -> 'b -> key -> key -> 'b
    val mapi_slice: (key -> 'a -> 'a) -> ('a t) -> key -> key -> ('a t)
    val forall_slice: (key -> 'a -> bool) -> ('a t) -> key -> key -> bool

  end

module Product (A: MAP_TYPE) (B: MAP_TYPE) =
  struct
    type key = [`A of A.key | `B of B.key]
(*    type (+'a) t *)
    type 'a t = ('a A.t) * ('a B.t)
    let empty = (A.empty, B.empty)
    let is_empty (a,b) = A.is_empty a && B.is_empty b
    let add key value (a, b) =
      match key with
	`A x -> (A.add x value a), b
      | `B y -> a, (B.add y value b)
    let find key (a, b) =
      match key with
	`A x -> A.find x a
      | `B x -> B.find x b
    let remove key (a, b) =
      match key with
	`A x -> (A.remove x a), b
      | `B y -> a, (B.remove y b)
    let mem key (a, b) =
      match key with
	`A x -> A.mem x a
      | `B y -> B.mem y b
    let iter f (a, b) =
      A.iter (fun x v -> f (`A x) v) a;
      B.iter (fun y v -> f (`B y) v) b
    let map f (a, b) =
      (A.map f a), (B.map f b)
    let mapi f (a, b) =
      (A.mapi (fun x v -> f (`A x) v) a),
      (B.mapi (fun y v -> f (`B y) v) b)
    let fold f (a, b) z =
      (A.fold (fun x v z -> f (`A x) v z) a
	 (B.fold (fun y v z -> f (`B y) v z) b z))
    let map2iz f (a1, b1) (a2, b2) =
      (A.map2iz (fun x v1 v2 -> f (`A x) v1 v2) a1 a2),
      (B.map2iz (fun y v1 v2 -> f (`B y) v1 v2) b1 b2)
    let map2i f1 f2 f (a1, b1) (a2, b2) =
      (A.map2i (fun x v1 -> f1 (`A x) v1)
	 (fun x v2 -> f2 (`A x) v2)
	 (fun x v1 v2 -> f (`A x) v1 v2) a1 a2),
      (B.map2i (fun y v1 -> f1 (`B y) v1)
	 (fun y v2 -> f2 (`B y) v2)
	 (fun y v1 v2 -> f (`B y) v1 v2) b1 b2)
    let map2io f1 f2 f (a1, b1) (a2, b2) =
      (A.map2io (fun x v1 -> f1 (`A x) v1)
	 (fun x v2 -> f2 (`A x) v2)
	 (fun x v1 v2 -> f (`A x) v1 v2) a1 a2),
      (B.map2io (fun y v1 -> f1 (`B y) v1)
	 (fun y v2 -> f2 (`B y) v2)
	 (fun y v1 v2 -> f (`B y) v1 v2) b1 b2)
    let map2 f1 f2 f (a1, b1) (a2, b2) =
      (A.map2 (fun x v1 -> f1 (`A x) v1)
	 (fun x v2 -> f2 (`A x) v2)
	 (fun x v1 v2 -> f (`A x) v1 v2) a1 a2),
      (B.map2 (fun y v1 -> f1 (`B y) v1)
	 (fun y v2 -> f2 (`B y) v2)
	 (fun y v1 v2 -> f (`B y) v1 v2) b1 b2)

   let iter2iz f (a1, b1) (a2, b2) =
     (A.iter2iz (fun x v1 v2 -> f (`A x) v1 v2) a1 a2);
     (B.iter2iz (fun x v1 v2 -> f (`B x) v1 v2) b1 b2)

   let iter2i f1 f2 f (a1, b1) (a2, b2) =
      (A.iter2i (fun x v1 -> f1 (`A x) v1)
	 (fun x v2 -> f2 (`A x) v2)
	 (fun x v1 v2 -> f (`A x) v1 v2) a1 a2);
      (B.iter2i (fun y v1 -> f1 (`B y) v1)
	 (fun y v2 -> f2 (`B y) v2)
	 (fun y v1 v2 -> f (`B y) v1 v2) b1 b2)

   let iter2 f1 f2 f (a1, b1) (a2, b2) =
      (A.iter2 (fun x v1 -> f1 (`A x) v1)
	 (fun x v2 -> f2 (`A x) v2)
	 (fun x v1 v2 -> f (`A x) v1 v2) a1 a2);
      (B.iter2 (fun y v1 -> f1 (`B y) v1)
	 (fun y v2 -> f2 (`B y) v2)
	 (fun y v1 v2 -> f (`B y) v1 v2) b1 b2)

    let forall2iz f (a1, b1) (a2, b2) =
      (A.forall2iz (fun x v1 v2 -> f (`A x) v1 v2) a1 a2) &&
      (B.forall2iz (fun y v1 v2 -> f (`B y) v1 v2) b1 b2)

    let fold2 (f1 : key -> 'a -> 'z -> 'z)
	(f2 : key -> 'b -> 'z -> 'z)
	(f : key -> 'a -> 'b -> 'z -> 'z)
	((a1, b1) : ('a t))
	((a2, b2) : ('b t)) (z0 : 'z) =
      A.fold2
	(fun x v1 z ->
	  f1 (`A x) v1 z)
	(fun x v2 z ->
	  f2 (`A x) v2 z)
	(fun x v1 v2 z ->
	  f (`A x) v1 v2 z)
	a1 a2
	(B.fold2
	   (fun y v1 z ->
	     f1 (`B y) v1 z)
	   (fun y v2 z ->
	     f2 (`B y) v2 z)
	   (fun y v1 v2 z ->
	     f (`B y) v1 v2 z)
	   b1 b2 z0)

    let fold2i
	(f : key -> 'a -> 'b -> 'z -> 'z)
	((a1, b1) : ('a t))
	((a2, b2) : ('b t)) (z0 : 'z) =
      A.fold2i
	(fun x v1 v2 z ->
	  f (`A x) v1 v2 z)
	a1 a2
	(B.fold2i
	   (fun y v1 v2 z ->
	     f (`B y) v1 v2 z)
	   b1 b2 z0)

    type 'a delta_record = ('a A.delta_record) * ('a B.delta_record)
    let compute_delta f (a1, b1) (a2, b2) =
      (A.compute_delta f a1 a2),
      (B.compute_delta f b1 b2)
    let apply_delta (a1, b1) (a2, b2) =
      (A.apply_delta a1 a2),
      (B.apply_delta b1 b2)

    (* applies the function to all the tuples *)
    (*let map_two f (a, b) =
      (A.map_two (fun x v -> f (`A x) v) a),
      (B.map_two (fun y v -> f (`B y) v) b)
    let fold_two f (a, b) z =
      (A.fold_two (fun x v z -> f (`A x) v  z) a
	 (B.fold_two (fun y v z -> f (`B y) v  z) b z))
    let iter_two f (a, b) =
      (A.iter_two f a); (B.iter_two f b)
    *)
    let equal f (a,b) (c,d) = 
	A.equal f a c && B.equal f b d 

    let size (a, b) = (A.size a) + (B.size b)
  end

